import json
import logging
import secrets
import time
from typing import List

from injector import inject
from paho.mqtt import client as mqtt_client
from paho.mqtt.enums import CallbackAPIVersion

from .client import StoreClient
from .. import config

log = logging.getLogger(__name__)


class QueueStore(StoreClient):
    @inject
    def __init__(self, configuration: config.Config):
        self.qos = configuration.queue_qos
        self.mqtt_prefix = configuration.queue_prefix

        client_id = f'klimalogger-mqtt-{configuration.client_host_name}-{secrets.randbelow(1000)}'

        def on_connect(client, userdata, flags, reason_code, properties):
            if reason_code == 0:
                log.debug("Connected to MQTT Broker!")
            else:
                log.error("Failed to connect, return code %d", reason_code)

        def on_disconnect(client, userdata, reason_code, properties):
            log.warning("Disconnected from MQTT Broker: %d", reason_code)

            delay = 5
            max_delay = 60

            while True:
                try:
                    if not client.reconnect():
                        log.info("Successful reconnect to MQTT Broker")
                        break
                except ConnectionRefusedError:
                    log.info("Reconnect to MQTT Broker failed, retry in %d seconds", delay)

                if delay < max_delay:
                    delay *= 2 + secrets.randbelow(5)

                time.sleep(delay)

        try:
            self.client = mqtt_client.Client(client_id=client_id, clean_session=False, callback_api_version = CallbackAPIVersion.VERSION2)
            # client.username_pw_set(username, password)
            self.client.on_connect = on_connect
            self.client.on_disconnect = on_disconnect
            log.info("connect to host %s, port %d", configuration.service_host, configuration.service_port)
            self.client.connect(configuration.service_host, configuration.service_port)
            self.client.loop_start()
        except Exception:
            log.exception("could not create client")
            self.client = None

    def __del__(self):
        if self.client:
            self.client.disconnect()

    def store(self, data: List[dict]):
        if self.client:

            if not self.client.is_connected():
                log.warning("client not connected, try to reconnect")
                self.client.reconnect()

            for entry in data:
                topic, json_message = self.map_entry(entry)
                message = json.dumps(json_message)
                log.info("write data (%d bytes) to topic %s", len(message), topic)
                result = self.client.publish(topic, payload=message, qos=self.qos, )
                status = result[0]
                if status != 0:
                    log.warning("Failed to send message to topic %s with status %s", topic, status)
        else:
            log.warning("client not available")
            raise RuntimeError("bla")

    def map_entry(self, entry: dict):
        timestamp = entry["time"]
        value = entry["fields"]["value"]
        tags = entry["tags"]
        measurement_type = tags["type"]
        unit = tags["unit"]
        sensor = tags["sensor"]
        topic = f"{self.mqtt_prefix}/{measurement_type}"
        print(f"{topic}: {value} {unit} ({sensor})")
        return (topic, {
            "time": int(timestamp),
            "value": value,
            "unit": unit,
            "sensor": sensor,
            "calculated": tags["calculated"]
        })



import math


class TemperatureCalc:

    def dew_point(self, temperature: float, relative_humidity: float):
        """
        Dew point calculation using the calculation published by Mark G. Lawrence (https://journals.ametsoc.org/view/journals/bams/86/2/bams-86-2-225.xml)

        :param temperature: Temperature in degrees Celsius
        :param relative_humidity: Relavite humidity in percent (range 0..1)

        :return: Calculated dew point in degrees Celsius
        """

        # Parameters according to Alduchov and Eskridge (https://journals.ametsoc.org/view/journals/apme/35/4/1520-0450_1996_035_0601_imfaos_2_0_co_2.xml)
        a = 17.625
        b = 243.04 # °C

        # Formula according to Lawrence:
        # Ts = (b × alpha(T,RH)) / (a - alpha(T,RH))
        # with alpha(T,RH) = ln(RH/100) + aT/(b+T)

        alpha = math.log(relative_humidity/100) + a * temperature / (b + temperature)

        return (b * alpha) / (a - alpha)



import configparser
import importlib
import logging
import time
from dataclasses import dataclass
from typing import Optional

from injector import singleton, inject, Injector
from lazy import lazy

from .data_builder import DataBuilderFactory, DataBuilder

log = logging.getLogger(__name__)


@dataclass
class Measurements:
    temperature: Optional[float] = None
    relative_humidity: Optional[float] = None


@singleton
class MeasurementDispatcher:
    @inject
    def __init__(self, configuration: configparser.ConfigParser, sensor_factory: "SensorFactory",
                 data_builder_factory: DataBuilderFactory):
        self.sensor_factory = sensor_factory
        self.data_builder_factory = data_builder_factory
        self.sensor_names = [sensor.strip() for sensor in configuration.get('client', 'sensors').split(',') if sensor.strip() != ""]

    def measure(self) -> DataBuilder:
        log.info("measure()")
        data_builder = self.data_builder_factory.get()
        measurements = Measurements()
        for sensor in self.sensors:
            try:
                start_time = time.time()
                sensor.measure(data_builder, measurements)
                end_time = time.time()
                measurement_duration = round((end_time - start_time) * 1e3, 1)
                data_builder.add(sensor.name, "time", "ms", measurement_duration)
            except Exception:
                log.exception("measurement of sensor %s failed", sensor)

        return data_builder

    @lazy
    def sensors(self):
        log.info("sensors()")
        sensors = [self.sensor_factory.create_sensor(sensor_name) for sensor_name in self.sensor_names]
        return sorted(sensors, key=lambda entry: entry.priority)


@singleton
class SensorFactory:

    @inject
    def __init__(self, current_injector: Injector):
        self.injector = current_injector

    def create_sensor(self, sensor_type: str):
        try:
            module = importlib.import_module('klimalogger.sensor.' + sensor_type + '_sensor')
            log.info("sensor: %s, module: %s", sensor_type, module)
            return self.injector.get(module.Sensor)
        except Exception:
            log.exception("instatiation of sensor %s failed", sensor_type)



import logging
import time

from injector import singleton, inject, Injector

from .config import ConfigModule, Config
from .data_builder import DataBuilder
from .measurement import MeasurementDispatcher
from .sensor import SensorModule
from .store import StoreClient
from .store import StoreModule
from . import logger

INJECTOR = Injector(
    [StoreModule(), ConfigModule(), SensorModule()])

root_logger = logging.getLogger(__name__)
root_logger.setLevel(logging.WARN)

log = root_logger


def set_log_level(log_level):
    root_logger.setLevel(log_level)


def add_log_handler(log_handler):
    root_logger.addHandler(log_handler)


def set_parent_logger(logger):
    logger.parent = root_logger


@singleton
class Client:
    @inject
    def __init__(self, measurement_dispatcher: MeasurementDispatcher,
                 store_client: StoreClient,
                 config: Config):
        self.measurement_dispatcher = measurement_dispatcher
        self.store_client = store_client
        self.config = config

    def measure_and_store_periodically(self, period=15):
        log.info("measure_and_store_periodically(%d)", period)
        last_measurement = time.time() - period
        skip_initial = 2
        while True:
            timestamp = time.time()
            if last_measurement < timestamp - period:
                if skip_initial == 0:
                    self.measure_and_store()
                else:
                    log.info("skipping initial measurement(%d)", skip_initial)
                    self.measure()
                    skip_initial -= 1
                last_measurement += period
            else:
                self.measure()
            log.info("time taken: %d", time.time() - last_measurement)
            time.sleep(1)

    def measure_and_store(self):
        log.info("measure_and_store()", )
        (timestamp, data) = self.measure()

        self.store_data(data, timestamp)

    def measure(self):
        result = self.measurement_dispatcher.measure()

        return result.timestamp, result.data

    def store_data(self, data, timestamp):
        try:
            self.store_client.store(data)
            log.info("stored data")
        except Exception as e:
            log.error("error during data transmission: create local log entry", exc_info=e)


def client():
    return INJECTOR.get(Client)



import logging
import os
from datetime import datetime
from typing import List

from .client import StoreClient

log = logging.getLogger(__name__)


class FileStore(StoreClient):
    target_folder = "/var/lib/klimalogger/data"

    def store(self, data: List[dict]):
        for entry in data:
            timestamp = datetime.fromtimestamp(entry["time"])
            value = entry["fields"]["value"]
            tags = entry["tags"]
            measurement_type = tags["type"]
            unit = tags["unit"]
            sensor = tags["sensor"]
            folder_name = f"{self.target_folder}/{timestamp:%Y}/{timestamp:%m}/"
            file_name = f"{measurement_type}_{timestamp:%Y%m%d}.txt"

            os.makedirs(folder_name, exist_ok=True)
            with open(folder_name + '/' + file_name, 'a') as output_file:
                output_file.write(f"{timestamp.isoformat()} {value} {unit} {sensor}\n")



# -*- coding: utf8 -*-

import time

import adafruit_scd4x
import busio
from injector import singleton, inject

from klimalogger import DataBuilder
from klimalogger.measurement import Measurements


@singleton
class Sensor:
    name = "SCD4x"
    priority = 3

    @inject
    def __init__(self, i2c_bus: busio.I2C):
        self.driver = adafruit_scd4x.SCD4X(i2c_bus)
        self.driver.start_periodic_measurement()

    def __del__(self):
        self.driver.stop_periodic_measurement()

    def measure(self, data_builder: DataBuilder, measurements: Measurements) -> None:

        while True:
            CO2 = self.driver.CO2

            if CO2 is not None:
                break

            time.sleep(1)

        data_builder.add(self.name, "CO2", "ppm", float(CO2))



# -*- coding: utf8 -*-
import logging

import busio
from injector import singleton, inject

from .. import DataBuilder
from ..calc import PressureCalc, TemperatureCalc
from ..measurement import Measurements

try:
    import configparser
except ImportError:
    import configparser as configparser

import adafruit_bme680

log = logging.getLogger(__name__)


@singleton
class Sensor:
    name = "BME680"
    priority = 1

    @inject
    def __init__(self, i2c_bus: busio.I2C, config_parser: configparser.ConfigParser, temperature_calc: TemperatureCalc,
                 pressure_calc: PressureCalc):
        log.info("init()")
        self.temperature_calc = temperature_calc
        self.pressure_calc = pressure_calc
        self.elevation = int(config_parser.get('bme680_sensor', 'elevation'))
        self.driver = adafruit_bme680.Adafruit_BME680_I2C(i2c_bus)
        self.driver.set_gas_heater(None, None)

    def measure(self, data_builder: DataBuilder, measurements: Measurements) -> None:
        temperature = self.driver.temperature
        relative_humidity = self.driver.relative_humidity
        dew_point = self.temperature_calc.dew_point(temperature, relative_humidity)
        pressure = self.driver.pressure
        sea_level_pressure = self.pressure_calc.sea_level_pressure(pressure, temperature, self.elevation)
        voc_gas = self.driver.gas

        data_builder.add(self.name, "temperature", "°C", round(temperature, 2))
        data_builder.add(self.name, "relative humidity", "%", round(relative_humidity, 2))
        data_builder.add(self.name, "dew point", "°C", round(dew_point, 2), is_calculated=True)
        data_builder.add(self.name, "pressure", "hPa", round(pressure, 2))
        data_builder.add(self.name, "sea level pressure", "hPa", round(sea_level_pressure, 2))
        data_builder.add(self.name, "voc gas", "Ohm", float(voc_gas))

        measurements.temperature = temperature
        measurements.relative_humidity = relative_humidity



# -*- coding: utf8 -*-

from injector import singleton, inject

from klimalogger import DataBuilder
from klimalogger.measurement import Measurements

try:
    import configparser
except ImportError:
    import configparser as configparser

import board
import busio
import adafruit_sgp30


@singleton
class Sensor:
    name = "SGP30"
    priority = 3

    @inject
    def __init__(self, i2c_bus: busio.I2C, config_parser: configparser.ConfigParser):
        self.baseline_eCO2 = int(config_parser.get('sgp30_sensor', 'baseline_eCO2'))
        self.baseline_TVOC = int(config_parser.get('sgp30_sensor', 'baseline_TVOC'))

        self.driver = adafruit_sgp30.Adafruit_SGP30(i2c_bus)

    def measure(self, data_builder: DataBuilder, measurements: Measurements) -> None:
        # self.sensor.set_iaq_baseline(self.baseline_eCO2, self.baseline_TVOC)
        eCO2, TVOC = self.driver.iaq_measure()

        data_builder.add(self.name, "eCO2", "ppm", float(eCO2))
        data_builder.add(self.name, "TVOC", "ppb", float(TVOC))



# -*- coding: utf8 -*-

import adafruit_sgp40
import busio
from injector import singleton, inject

from klimalogger import DataBuilder
from klimalogger.measurement import Measurements


@singleton
class Sensor:
    name = "SGP40"
    priority = 3

    @inject
    def __init__(self, i2c_bus: busio.I2C):
        self.driver = adafruit_sgp40.SGP40(i2c_bus)

    def measure(self, data_builder: DataBuilder, measurements: Measurements) -> None:
        if measurements.temperature is not None and measurements.relative_humidity is not None:
            data_builder.add(self.name, "VOC index", "", float(
                self.driver.measure_index(
                    temperature=measurements.temperature,
                    relative_humidity=measurements.relative_humidity
                )))
        else:
            data_builder.add(self.name, "raw gas", "", float(self.driver.raw))



class PressureCalc:

    def sea_level_pressure(self, pressure: float, temperature_celsius: float, elevation: float) -> float:
        temperature = temperature_celsius + 273.15
        return pressure * pow(temperature / (temperature + 0.0065 * elevation), -5.255)



import busio
from injector import Module, singleton, provider


class SensorModule(Module):
    @singleton
    @provider
    def provide_i2c_bus(self) -> busio.I2C:
        import board
        return busio.I2C(board.SCL, board.SDA, frequency=100000)



#!/usr/bin/env python3

import secrets
import sys

import paho.mqtt.client as mqtt_client


def connect_mqtt(broker: str):
    port = 1883
    client_id = f'python-mqtt-{secrets.randbelow(1000)}'

    def on_connect(client, userdata, flags, reason_code, properties):
        if reason_code == 0:
            print("Connected to MQTT Broker!")
        else:
            print("Failed to connect, return code %d\n", reason_code)

    # Set Connecting Client ID
    client = mqtt_client.Client(client_id=client_id, callback_api_version=mqtt_client.CallbackAPIVersion.VERSION2)
    # client.username_pw_set(username, password)
    client.on_connect = on_connect
    client.connect(broker, port)
    return client


def subscribe(client: mqtt_client.Client, topic: str):
    def on_message(client, userdata, msg):
        print(f"Received `{msg.payload.decode()}` from `{msg.topic}` topic")

    client.subscribe(topic)
    client.on_message = on_message


def main():
    topic = "sensors/#"

    broker = sys.argv[1]
    client = connect_mqtt(broker)
    subscribe(client, topic)
    client.loop_forever()


if __name__ == "__main__":
    main()
